#!/usr/bin/env python

# jahplayer - A widget library representation.

# Copyright (C) 2006 VM Inc.
# Released under the GPL.
# For more information, see http://www.jahshaka.org.

# This is required for the executable generated from py2exe
# to use the preinstalled openlibraries python bindings

import bootstrap
import sys
import os

import platform
import os.path
import pickle
import jahwidgets
import utilities
import openpluginlib
import openassetlib
import openmedialib
import uuid
import rpn

#
# Define guid generator class acceptable to oal
#

class uuid_generator( openassetlib.guid_generator ):
	def generate_guid( self ):
		return uuid.uuid_time( )

#
# The encoder widget
#

class encoder:
    def __init__( self, app ):
	# some useful default property sets
	self.sar_keys = [ "Default", "1:1", "NTSC 4:3", "NTSC 16:9", "PAL 4:3", "PAL 16:9" ]
	self.sar_modes = { }
	self.sar_modes[ self.sar_keys[ 0 ] ] = ( 0, 0 )
	self.sar_modes[ self.sar_keys[ 1 ] ] = ( 1, 1 )
	self.sar_modes[ self.sar_keys[ 2 ] ] = ( 10, 11 )
	self.sar_modes[ self.sar_keys[ 3 ] ] = ( 40, 33 )
	self.sar_modes[ self.sar_keys[ 4 ] ] = ( 59, 54 )
	self.sar_modes[ self.sar_keys[ 5 ] ] = ( 118, 81 )

	self.rescale_keys = [ "320x240", "640x480", "720x576", "1024x768", "1280x1024", "1440x1080", "1600x1200" ]
	self.rescale_modes = { }
	self.rescale_modes[ self.rescale_keys[ 0 ] ] = ( 320, 240 )
	self.rescale_modes[ self.rescale_keys[ 1 ] ] = ( 640, 480 )
	self.rescale_modes[ self.rescale_keys[ 2 ] ] = ( 720, 576 )
	self.rescale_modes[ self.rescale_keys[ 3 ] ] = ( 1024, 768 )
	self.rescale_modes[ self.rescale_keys[ 4 ] ] = ( 1280, 1024 )
	self.rescale_modes[ self.rescale_keys[ 5 ] ] = ( 1440, 1080 )
	self.rescale_modes[ self.rescale_keys[ 6 ] ] = ( 1600, 1200 )

	self.fps_keys = [ "Default", "12 fps", "15 fps", "24 fps", "PAL", "NTSC", "30 fps" ]
	self.fps_modes = { }
	self.fps_modes[ self.fps_keys[ 0 ] ] = ( 25, 1 )
	self.fps_modes[ self.fps_keys[ 1 ] ] = ( 12, 1 )
	self.fps_modes[ self.fps_keys[ 2 ] ] = ( 15, 1 )
	self.fps_modes[ self.fps_keys[ 3 ] ] = ( 24, 1 )
	self.fps_modes[ self.fps_keys[ 4 ] ] = ( 25, 1 )
	self.fps_modes[ self.fps_keys[ 5 ] ] = ( 30000, 1001 )
	self.fps_modes[ self.fps_keys[ 6 ] ] = ( 30, 1 )

	self.audio_keys = [ "22050", "32000", "44100", "48000" ]
	self.audio_modes = {}
	self.audio_modes[ self.audio_keys[ 0 ] ] = 22050
	self.audio_modes[ self.audio_keys[ 1 ] ] = 32000
	self.audio_modes[ self.audio_keys[ 2 ] ] = 44100
	self.audio_modes[ self.audio_keys[ 3 ] ] = 48000

	self.dv_keys = [ "PAL", "PAL 16:9", "NTSC", "NTSC 16:9" ]
	self.dv_modes = {}
	self.dv_modes[ self.dv_keys[ 0 ] ] = { "rescale": (720, 576), "fps": (25, 1), "sar": (59, 54), "aspect_ratio": 4.0 / 3.0 }
	self.dv_modes[ self.dv_keys[ 1 ] ] = { "rescale": (720, 576), "fps": (25, 1), "sar": (118, 81), "aspect_ratio": 16.0 / 9.0 }
	self.dv_modes[ self.dv_keys[ 2 ] ] = { "rescale": (720, 480), "fps": (30000, 1001), "sar": (10, 11), "aspect_ratio": 4.0 / 3.0 }
	self.dv_modes[ self.dv_keys[ 3 ] ] = { "rescale": (720, 480), "fps": (30000, 1001), "sar": (40, 33), "aspect_ratio": 16.0 / 9.0 }

	# setup the ui
	self.app = app
	self.handle = jahwidgets.create_ui_from_file( utilities.get_absolute_path( "encoder.ui" ), jahwidgets.widget_handle() )
	
	self.encode_prop = utilities.create_property_and_attach( self.handle, "pb_encode:clicked", utilities.call_method_observer( self.encode ) )
	self.cancel_prop = utilities.create_property_and_attach( self.handle, "pb_cancel:clicked", utilities.call_method_observer( self.cancel ) )
	self.progress_label_prop = utilities.create_property( self.handle, "label_progress:text" )
	self.progress_prop = utilities.create_property( self.handle, "progress:progress" )
	self.progress_total_prop = utilities.create_property( self.handle, "progress:total" )
	self.types_prop = utilities.create_property_and_attach( self.handle, "encoder_type:currentItem", utilities.call_method_observer( self.type_changed ) )
	self.filename_prop = utilities.create_property_and_attach( self.handle, "filename:text", utilities.call_method_observer( self.set_filename ) )
	self.dir_chooser_prop = utilities.create_property_and_attach( self.handle, "dir_chooser:clicked", utilities.call_method_observer( self.choose_dir ) )

	# initialize filename
	self.filename = "output"
	self.dir = os.getcwd()
	self.filename_prop.set( self.filename )

	# initialize progress bar
	self.progress_total_prop.set( 100.0 )
	self.progress_prop.set( 0.0 )
	self.progress_label_prop.set( "idle" )

	# populate known types
	self.encoder_types = [ "mjpeg", "dv" ]
	jahwidgets.set_value( "encoder_type:items", self.handle, openpluginlib.any( self.encoder_types ) )

	self.encoder_config = {}
	self.encoder_config[ "mjpeg" ] = { "ui": "mjpeg.ui",
					   "hookup": self.mjpeg_hookup,
					   "encode": self.mjpeg_encode
					   }
	self.encoder_config[ "dv" ] = { "ui": "dv.ui",
					"hookup": self.dv_hookup,
					"encode": self.dv_encode
					}

	self.encoder_ui_handle = None
	self.type_changed( self.types_prop )
	self.encoder_running = False

    def set_filename( self, prop ):
	self.filename = self.filename_prop.value_as_wstring()

    def choose_dir( self, prop ):
	self.dir = jahwidgets.get_directory( self.dir, "choose output directory" )

    def get_output_filename( self, extension ):
	# force extension to be correct
	if not self.filename.rfind('.') == -1:
	    self.filename = self.filename[:self.filename.rfind('.')]
	    
	result = os.path.join( self.dir, self.filename ) + "." + extension
	return str( result )

    def show( self ):
	# can also be considered a reset point
	self.progress_label_prop.set( "idle" )
	self.handle.show( True )

    def set_media_information( self, media_list ):
	self.media_list = media_list;

    def encode( self, property ):
	self.should_quit = False
	self.encoder_running = True
	
	self.progress_prop.set( 0.0 )
	self.progress_label_prop.set( "encoding" )

	enc_type = self.encoder_types[ self.types_prop.value_as_int() ]
	self.encoder_config[ enc_type ][ "encode" ]()

	self.progress_prop.set( 0.0 )
	self.progress_label_prop.set( "done" )

	self.encoder_running = False
	
    def cancel( self, property ):
	if self.encoder_running:
	    self.should_quit = True
	    return

	jahwidgets.application().quit()

    #
    # callbacks used from within rpn
    #
    def query_cancel( self ):
	return self.should_quit

    def update_progress( self, v ):
	self.progress_prop.set( v )
	self.app.processEvents( 50 )

    def init_progress( self, v ):
	self.progress_total_prop.set( float( v ) )

    def type_changed( self, property ):
	if not self.encoder_ui_handle == None:
	    self.encoder_ui_handle.remove()
	    
	enc_type = self.encoder_types[ self.types_prop.value_as_int() ]
	self.encoder_ui_handle = jahwidgets.create_ui_from_file( utilities.get_absolute_path( self.encoder_config[ enc_type ][ "ui" ] ),
								 jahwidgets.find( "parameter_frame", self.handle ) )
	self.encoder_config[ enc_type ][ "hookup" ]()
	self.encoder_ui_handle.show( True )

    def construct_input_sub_filter( self, input, _in, out, fps, audio ):
	return [ input,
				 "filter:conform",
		 "filter:resampler", "frequency=%i" % ( audio ),
		 "filter:clip", "in=%i" % ( _in ), "out=%i" % ( out ),
		 #"filter:threader",
		 "filter:wrapper", "fps_num=%i" % ( fps[0] ), "fps_den=%i" % ( fps[1] ) ]

    def construct_filter_graph_description( self, sar, fps, rescale, audio ):
	graph = [ "colour:",
		  "width=%i" % ( rescale[0] ),
		  "height=%i" % ( rescale[1] ),
		  "sar_num=%i" % ( sar[0] ),
		  "sar_den=%i" % ( sar[1] ),
		  "fps_num=%i" % ( fps[0] ),
		  "fps_den=%i" % ( fps[1] ) ]
	
	for m in self.media_list:
	    graph.extend( self.construct_input_sub_filter( m[ "filename" ], m[ "in" ], m[ "out" ], fps, audio ) )

	graph.extend( [ "filter:playlist", "slots=%i" % ( len( self.media_list ) ),
			"filter:deinterlace",
			"filter:composite", "@0.out=@1.rpn_length" ] )
	return graph

    # mjpeg methods
    def mjpeg_hookup( self ):
	jahwidgets.set_value( "fps:items", self.handle, openpluginlib.any( self.fps_keys ) )
	jahwidgets.set_value( "rescale:items", self.handle, openpluginlib.any( self.rescale_keys[:3] ) )
	jahwidgets.set_value( "audio:items", self.handle, openpluginlib.any( self.audio_keys ) )

    def mjpeg_encode( self ):
	fps = self.fps_modes[ self.fps_keys[ jahwidgets.get_value( "fps:currentItem", self.handle ).as_int() ] ]
	rescale = self.rescale_modes[ self.rescale_keys[ jahwidgets.get_value( "rescale:currentItem", self.handle ).as_int() ] ]
	audio = self.audio_modes[ self.audio_keys[ jahwidgets.get_value( "audio:currentItem", self.handle ).as_int() ] ]

	graph = self.construct_filter_graph_description( ( 1, 1 ), fps, rescale, audio )

	output = self.get_output_filename( "avi" )
	store_props = { "name": output, "vcodec": "mjpeg", "aspect_ratio": "1.0" }
	
	rpn.rpn( graph,
		 store_props,
		 self.query_cancel,
		 self.init_progress,
		 self.update_progress )

    # dv methods
    def dv_hookup( self ):
	jahwidgets.set_value( "format:items", self.handle, openpluginlib.any( self.dv_keys ) )

    def dv_encode( self ):
	f = self.dv_modes[ self.dv_keys[ jahwidgets.get_value( "format:currentItem", self.handle ).as_int() ] ]	       
	graph = self.construct_filter_graph_description( f[ "sar" ], f[ "fps" ], f[ "rescale" ], 48000 )

	output = self.get_output_filename( "dv" )
	store_props = { "name": output, "vcodec": "dv", "aspect_ratio": str( f[ "aspect_ratio" ] ) }
	
	rpn.rpn( graph,
		 store_props,
		 self.query_cancel,
		 self.init_progress,
		 self.update_progress )

#
# Run the main method if required
#

if __name__ == "__main__":
  # Turn on trace if required
	if len( sys.argv ) > 1 and sys.argv[ 1 ] == "--trace":
		sys.settrace( utilities.traceit )
		sys.argv.pop( 1 )

	print "encoder: " + str(sys.argv)
	
	sys.argv.pop( 0 )

	# Initialise the application
	app = jahwidgets.application( )
    
	# setup some styling
	if not app.setStyle( "v3" ):
		print "failed to load v3 theme"
 
	# Initialise the open libraries
	openassetlib.init( uuid_generator( ) )

	enc = encoder( app )
	if not len( sys.argv ) == 0:
		enc.set_media_information( pickle.load( open( sys.argv[ 0 ], "r" ) ) )
		enc.show()

		# Execute the application
		app.start_event_loop( )

# Close opl
openassetlib.uninit( )
