import openmedialib
import openpluginlib
import balance
import common

def assign_prop( property, value ):
	if property is not None and property.valid( ):
		if property.is_a_int( ):
			property.set( int( value ) )
		elif property.is_a_double( ):
			property.set( float( value ) )
		elif property.is_a_wstring( ):
			property.set( str( value ) )
		elif property.is_a_wstring_list( ):
			property.set( value )
	else:
		print "invalid ", property, value

def value_prop( prop ):
	result = None
	if prop is not None and prop.valid( ):
		if prop.is_a_int( ):
			result = str( prop.value_as_int( ) )
		elif prop.is_a_double( ):
			result = str( prop.value_as_double( ) )
		elif prop.is_a_wstring( ):
			result = unicode( prop.value_as_wstring( ) )
		elif prop.is_a_wstring_list( ):
			result = prop.value_as_wstring_list( )
	return result

def find_prop( node, key ):
	if key[ 0 ] == "@":
		split = key[ 1: ].split( "." )
		name = split.pop( )
		for i in split:
			node = node.fetch_slot( int( i ) )
			if node is None: return None
	else:
		name = key
	return node.properties( ).get_property( name )

def update_prop( node, name, value ):
	if node is not None:
		prop = node.properties( ).get_property( name )
		if prop is None or not prop.valid( ):
			prop = openpluginlib.property( openpluginlib.create_key_from_string( name ) )
			node.properties( ).append( prop )
		prop.set( value )

def retrieve_uri( node ):
	uri = None
	if node is not None:
		prop = find_prop( node, "rpn_uri" )
		if prop is not None and prop.valid( ):
			uri = value_prop( prop )
		else:
			uri = node.get_uri( )
	return uri

class operator:
	def __init__( self, uri, input = None, callback = None ):
		self.uri = uri
		self.properties = []
		self.input = input
		self.callback = callback

	def __str__( self ):
		return self.uri + " " + str( self.properties )

	def assign( self, node ):
		for i in self.properties:
			split = i.split( "=" )
			property = find_prop( node, split[ 0 ] )
			if split[ 1 ][ 0 ] == "@":
				temp = find_prop( node, split[ 1 ] )
				value = value_prop( temp )
			else:
				value = split[ 1 ]
			if property is not None:
				assign_prop( property, value )

class project:
	def __init__( self ):
		self.objects = { }

	def withdraw( self, op ):
		result = None
		if op.input is None:
			uri = op.uri
			if not self.objects.has_key( uri ):
				self.objects[ uri ] = [ ]
			instances = self.objects[ uri ]
			if len( instances ) > 0:
				result = instances.pop( )
			elif uri == "filter:wrapper":
				result = common.input_wrapper( )
			elif uri.find( "filter:" ) == 0:
				result = openmedialib.create_filter( str( uri[ 7: ] ) )
			else:
				result = openmedialib.create_input( str( uri ) )
			update_prop( result, "rpn_uri", str( uri ) )
		else:
			result = op.input
		return result

	def deposit( self, input ):
		if input is not None:
			i = 0
			while i < input.slots( ):
				self.deposit( input.fetch_slot( i ) )
				input.connect( None, i )
				i += 1
			uri = retrieve_uri( input )
			update_prop( input, "rpn_uri", str( uri ) )
			self.objects[ uri ].append( input )

class stack:
	def __init__( self, project ):
		self.project = project
		self.stack = [ ]
		self.dict = { "swap": self.swap, "2swap": self.swap2, "rot": self.rot, 
					  "drop": self.drop, "2drop": self.drop2, "dup": self.dup, 
					  "debug": self.debug, "decap": self.decap, "clone": self.clone,
					  "over": self.over, "render": self.render, "nip": self.nip,
					  "tuck": self.tuck, "!": self.assign }
		self.vars = { }

	def swap( self ):
		tos = self.pop( )
		bos = self.pop( )
		self.push_input( tos )
		self.push_input( bos )

	def swap2( self ):
		i1 = self.pop( )
		i2 = self.pop( )
		i3 = self.pop( )
		i4 = self.pop( )
		self.push_input( i2 )
		self.push_input( i1 )
		self.push_input( i4 )
		self.push_input( i3 )

	def rot( self ):
		tos = self.pop( )
		bos = self.pop( )
		uos = self.pop( )
		self.push_input( bos )
		self.push_input( tos )
		self.push_input( uos )

	def dup( self ):
		tos = self.pop( )
		self.push_input( tos )
		self.push_input( tos )

	def drop( self ):
		tos = self.pop( )
		self.project.deposit( tos )

	def drop2( self ):
		self.drop( )
		self.drop( )

	def decap( self ):
		tos = self.pop( )
		if tos is not None:
			i = 0
			while i < tos.slots( ):
				self.push_input( tos.fetch_slot( i ) )
				tos.connect( None, i )
				i += 1
			self.project.deposit( tos )

	def clone_node( self, node ):
		if node is not None:
			i = 0
			while i < node.slots( ):
				self.clone_node( node.fetch_slot( i ) )
				i += 1
			self.push( retrieve_uri( node ) )
			self.push_args( value_prop( find_prop( node, "rpn_props" ) ) )
			self.push_callback( node.fetch_callback( ) )

	def clone( self ):
		tos = self.pop( )
		self.push_input( tos )
		self.clone_node( tos )

	def over( self ):
		tos = self.pop( )
		bos = self.pop( )
		self.push_input( bos )
		self.push_input( tos )
		self.clone_node( bos )

	def nip( self ):
		self.swap( )
		self.drop( )

	def tuck( self ):
		self.clone( )
		tos = self.pop( )
		self.swap( )
		self.push_input( tos )

	def assign( self ):
		op = self.stack.pop( )
		tos = self.pop( )
		self.vars[ op.uri ] = tos

	def push( self, id ):
		result = True
		if id in self.dict.keys( ):
			self.dict[ id ]( )
		elif id in self.vars.keys( ):
			self.push_input( self.vars[ id ] )
		elif id.find( "/sequence:" ) > 0:
			self.stack.append( operator( id ) )
		else:
			if id.find( "=" ) > 0:
				self.stack[ -1 ].properties.append( id )
			elif id != "":
				self.stack.append( operator( id ) )
			else:
				result = False
		return result

	def push_callback( self, callback ):
		if len( self.stack ) > 0:
			self.stack[ -1 ].callback = callback

	def push_args( self, properties ):
		for i in properties: self.push( i )

	def push_input( self, input ):
		id = retrieve_uri( input )
		self.stack.append( operator( id, input ) )
		self.push_args( value_prop( find_prop( input, "rpn_props" ) ) )
		self.push_callback( input.fetch_callback( ) )
		return True

	def pop( self ):
		result = None
		if len( self.stack ) > 0:
			op = self.stack.pop( )
			result = self.project.withdraw( op )
			if result is not None:
				op.assign( result )
				slot = result.slots( )
				while slot > 0:
					slot -= 1
					if result.fetch_slot( slot ) is not None:
						self.push_input( result.fetch_slot( slot ) )
					result.connect( self.pop( ), slot )
				result.register( op.callback )
				update_prop( result, "rpn_length", result.get_frames( ) )
				update_prop( result, "rpn_props", [] + op.properties )
				op.assign( result )
		return result

	def debug( self, title = "stack trace" ):
		print title
		for i in self.stack:
			print i
		print

	def render( self, title = "top of stack" ):
		print title
		tos = self.pop( )
		balance.walk( tos )
		self.push_input( tos )

